'use strict'

/*
http://frutbunn.tk  frutbunn@gmail.com (frutbunn)

This work is licensed under the Creative Commons Attribution 4.0 International License. 
To view a copy of this license, visit http://creativecommons.org/licenses/by/4.0/.
*/

// Example:
// cuddlyLevel = new FruRenderMap(resources.image.tiles, cuddlyMap, 16*2, state.ctx.game.canvas, 0, 0, 320*2, 144*2);

function FruRenderMap(tileImageOrCanvas, mapData, blockSize, displayCanvas, x, y, width, height) {
	var that = this;

	var state = {
		ctx: {
			map: undefined,
			mapPrevious: undefined,

			output: undefined
		},

		tileImageOrCanvas: undefined,
		tile: undefined,

		mapData: undefined,

		blockSize: {
			w: undefined,
			h: undefined
		},

		displayArea: {
			x: undefined,
			y: undefined,
			w: undefined,
			h: undefined
		},

		horizontalBlockPosition: undefined
	}

	//	-	-	-

	function isArray(x) { return x && typeof x === 'object' && x.constructor === Array; }

	function makeCanvasContext(width, height) {
		var canvas = document.createElement('canvas');
		canvas.width = width;
		canvas.height = height;
		
		var ctx = canvas.getContext("2d");	
		return ctx;
	}

	function constructor() {
		state.tileImageOrCanvas = tileImageOrCanvas;
		state.mapData = mapData;

		if (isArray(blockSize) ) {
			state.blockSize.w = blockSize[0];
			state.blockSize.h = blockSize[1];
		} else {
			state.blockSize.w = state.blockSize.h = blockSize;
		}
		state.ctx.output = displayCanvas.getContext("2d");

		state.displayArea.x = x || 0;
		state.displayArea.y = y || 0;
		state.displayArea.w = width || state.ctx.output.canvas.width;
		state.displayArea.h = height || state.ctx.output.canvas.height;		

		state.horizontalBlockPosition = 0;

		var w = state.displayArea.w + state.blockSize.w;
		var h = state.mapData.length*state.blockSize.h;
		state.ctx.map = makeCanvasContext(w, h);
		state.ctx.mapPrevious = makeCanvasContext(w, h);

		state.tile = new Tile(state.tileImageOrCanvas.width, state.blockSize.w, state.blockSize.h);

		fill();
	}

	//	-	-	-

	function Coordinate(x, y) {
		this.x = x || 0;
		this.y = y || 0;
	}

	function Bob(x, y, w, h) {
		this.x = x || 0;
		this.y = y || 0;
		this.w = w || 0;
		this.h = h || 0;
	}

	//	-	-	-

	function Tile(canvasWidth, tileWidth, tileHeight, horizontalTileCount) {
		var state = {
			horizontalTileCount: null,

			tileSize: {
				width: null,
				height: null
			}
		}

		function constructor() {
			if (horizontalTileCount)
				state.horizontalTileCount = horizontalTileCount;
			else
				state.horizontalTileCount = Math.floor(canvasWidth/tileWidth);

			state.tileSize.height = Number(tileHeight);
			state.tileSize.width = Number(tileWidth);
		}

		function calcTilePosition(index, horizontalTileCount) {
			var r = new Coordinate(Math.floor(index%horizontalTileCount), Math.floor(index/horizontalTileCount) );
			
			return r;
		}

		function position(index) {
			var b = calcTilePosition(index, state.horizontalTileCount);

			b.x*= state.tileSize.width;
			b.y*= state.tileSize.height;

			return new Bob(b.x, b.y, state.tileSize.width, state.tileSize.height);
		}

		constructor();

		return {
			get width() { return state.tileSize.width; },
			get height() { return state.tileSize.height; },

			position: position
		}
	}

	//	-	-	-

	function clipToBounds(posX, posY) {
		if (posX<0) posX = 0;
		if (posY<0) posY = 0;

		if (posX>=state.mapData[0].length*state.blockSize.w - state.displayArea.w)
			posX = state.mapData[0].length*state.blockSize.w - state.displayArea.w;

		if (posY>=state.mapData.length*state.blockSize.h - state.displayArea.h)
			posY = state.mapData.length*state.blockSize.h - state.displayArea.h;

		return new Coordinate(posX, posY);		
	}

	function render(posX, posY) {
		// Calc clipped rendering positon
		var location = clipToBounds(posX, posY);

		// Take into account any movement of map position
		scroll(location.x, location.y);

		// Redraw foreground on framebuffer
		var s = new Bob(location.x % state.blockSize.w, location.y, state.displayArea.w, state.displayArea.h);
		state.ctx.output.drawImage(
			state.ctx.map.canvas,
			s.x, s.y, s.w, s.h,
			state.displayArea.x, state.displayArea.y, s.w, s.h
		);	

	}

	function getHorizontalBlockPosition() {
		return state.horizontalBlockPosition;
	}

	function setHorizontalBlockPosition(posX) {
		state.horizontalBlockPosition = Math.floor(posX/state.blockSize.w );
	}

	function drawColumn(x) {
		var offset = getHorizontalBlockPosition();

		for(var y=0; y<state.mapData.length; y++) {
			var s = state.tile.position( state.mapData[y][offset+x] );

			state.ctx.map.drawImage(
				state.tileImageOrCanvas,
				s.x, s.y, s.w, s.h,
				x*state.blockSize.w, y*state.blockSize.h, s.w, s.h
			);

		}

	}

	function fill() {
		var offset = getHorizontalBlockPosition();

		state.ctx.map.clearRect(0, 0, state.ctx.map.canvas.width, state.ctx.map.canvas.height);

		for(var x=0; x<state.displayArea.w/state.blockSize.w + 1; x++)
			drawColumn(x);

	}

	function flipBuffers() {
		var tmp = state.ctx.map;
		state.ctx.map = state.ctx.mapPrevious;
		state.ctx.mapPrevious = tmp;
	}

	function move(current, previous) {
		// Calc direction and width of part of foreground that hasn't changed.
		var dx = 0, sx = 0, w = 0;
		var _MAX_WIDTH = state.ctx.map.canvas.width/state.blockSize.w;

		var displacement = previous-current;
		if (displacement<0) {
			dx = 0;
			sx = Math.abs(displacement);
		} else {
			dx = displacement;
			sx = 0;
		}
		w = _MAX_WIDTH - Math.abs(displacement);

		flipBuffers();

		state.ctx.map.globalCompositeOperation='copy';

		var s = new Bob(0, 0, w*state.blockSize.w, state.ctx.map.canvas.height);
		state.ctx.map.drawImage(
			state.ctx.mapPrevious.canvas,
			sx*state.blockSize.w, 0, s.w, s.h,
			dx*state.blockSize.w, 0, s.w, s.h
		);		

		state.ctx.map.globalCompositeOperation='source-over';
	}

	function scroll(posX, posY) {
		// Make a copy of previous horizontal block position of foreground, then
		// update current horizontal block position.
		var previous = getHorizontalBlockPosition();
		setHorizontalBlockPosition(posX);
		var current = getHorizontalBlockPosition();

		// If current horizontal block hasn't changed, no update to foreground needed,
		// so do nothing.
		if (previous == current)
			return;

		// If moved more than one screen width, just redraw the lot instead of trying to scroll.
		var widthInBlocks = state.ctx.map.canvas.width/state.blockSize.w - 1;
		if(Math.abs(current-previous)>widthInBlocks) {
			fill();

			return;
		}

		// block move foreground based on change, then draw in missing columns 
		move(current, previous);

		var columns = Math.abs(current-previous);
		for(var i=0; i<columns; i++)
			if (previous>current)
				drawColumn(0+i);
			else
				drawColumn(widthInBlocks-i);

	}

	//	-	-	-

	constructor();

	return {
		render: render,
		get widthInPixels() { return state.mapData[0].length*state.blockSize.w; },
		get heightInPixels() { return state.mapData.length*state.blockSize.h; } 
	}

}